#include "fields.h"
#include <time.h>
#include <stdio.h>
#include <curses.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/resource.h>
#include <signal.h>
#include <pwd.h>
#include "sorting.h"

static FieldDefinition field_def[]={
	{	5,5,5,		"Pid"		},
	{	4,8,8,		"User"		},
	{	3,3,3,		"Pri"		},
	{	3,3,3,		"NI"		},
	{	4,7,5,		"vsz"		},
	{	4,7,5,		"res"		},
	{	4,7,5,		"Share"		},
	{	4,4,4,		"Stat"		},
	{	3,6,6,		"%cpu"		},
	{	3,6,6,		"%mem"		},
	{	5,5,5,		"time"		},
	{	8,128,16,	"Command"	},
};
static const unsigned fields = 12;

static time_t grab_time;
static time_t uptime;
static unsigned user_count;
static double loadavg[3];

static unsigned proccount_total;
static unsigned proccount_sleep;
static unsigned proccount_running;
static unsigned proccount_stopped;
static unsigned proccount_zombie;

unsigned cpus;

struct CPU {
	unsigned id;
	bool online;
	double user;
	double sys;
	double nice;
	double idle;
};
static CPU *cpu=0;
static double total_cpu_user=44.4;
static double total_cpu_sys=2.3;
static double total_cpu_nice=0.1;
static double total_cpu_idle=52.2;

static unsigned mem_total_kb;
static unsigned mem_used_kb;
static unsigned mem_free_kb;
static unsigned swap_total_kb;
static unsigned swap_used_kb;
static unsigned swap_free_kb;

struct Process {
	pid_t		pid;
	uid_t		userid;
	int		priority;
	int		nice;
	unsigned	vsz_kb;
	unsigned	res_kb;
	unsigned	share_kb;
	unsigned	state;
	double		cpu_usage;
	bool		is_idle;
	double		mem_usage;
	time_t		start_time;
	time_t		cpu_time;
	char		command[64];
};
static Process *process=0;
static unsigned processes=0;
static unsigned process_alloc=0;


void grab_info(bool /*filer_idle*/, const char */*filter_userid*/) {
	grab_time = time(0);
	uptime = 0;
	user_count=1;
	loadavg[0]=0.57;
	loadavg[1]=1.37;
	loadavg[2]=2.01;
	
	cpus = 4;
	cpu = (CPU*)malloc(sizeof(CPU)*cpus);
	cpu[0].id = 10;
	cpu[0].online = true;
	cpu[0].user = 9.3;
	cpu[0].sys = 4.6;
	cpu[0].nice = 0.2;
	cpu[0].idle = 85.7;
	cpu[1].id = 11;
	cpu[1].online = false;
	cpu[1].user = 0.0;
	cpu[1].sys = 0.0;
	cpu[1].nice = 0.0;
	cpu[1].idle = 0.0;
	cpu[2].id = 20;
	cpu[2].online = true;
	cpu[2].user = 0.0;
	cpu[2].sys = 0.3;
	cpu[2].nice = 0.0;
	cpu[2].idle = 99.6;
	cpu[3].id = 21;
	cpu[3].online = true;
	cpu[3].user = 16.4;
	cpu[3].sys = 1.6;
	cpu[3].nice = 44.4;
	cpu[3].idle = 37.3;
	
	proccount_total = 80;
	proccount_sleep = 70;
	proccount_running = 1;
	proccount_stopped = 1;
	proccount_zombie = 8;
	
	mem_total_kb = 642196;
	mem_used_kb = 313660;
	mem_free_kb = 328536;
	swap_total_kb = 530136;
	swap_used_kb = 0;
	swap_free_kb = 530136;
	
	process = (Process*)malloc(sizeof(Process)*3);
	process_alloc=3;
	processes = 3;
	process[0].pid = 2116;
	process[0].userid = 500;
	process[0].priority = 20;
	process[0].nice = 19;
	process[0].vsz_kb = 278*1024;
	process[0].res_kb = 600;
	process[0].share_kb = 500;
	process[0].state = 0;
	process[0].cpu_usage = 97.2;
	process[0].is_idle = false;
	process[0].mem_usage = 0.0;
	process[0].cpu_time = 34597;
	strcpy(process[0].command,"md5crk");
	process[1].pid = 2639;
	process[1].userid = 511;
	process[1].priority = 11;
	process[1].nice = 0;
	process[1].vsz_kb = 1044;
	process[1].res_kb = 1040;
	process[1].share_kb = 824;
	process[1].state = 0;
	process[1].cpu_usage = 1.9;
	process[1].is_idle = true;
	process[1].mem_usage = 0.1;
	process[1].cpu_time = 0;
	strcpy(process[1].command,"top");
	process[2].pid = 1;
	process[2].userid = 1;
	process[2].priority = 8;
	process[2].nice = 0;
	process[2].vsz_kb = 9;
	process[2].res_kb = 508;
	process[2].share_kb = 508;
	process[2].state = 0;
	process[2].cpu_usage = 0.0;
	process[2].is_idle = false;
	process[2].mem_usage = 0.0;
	process[2].cpu_time = 4;
	strcpy(process[2].command,"init");
}

void paint_top(WINDOW *w) {
	wclear(w);
	
	struct tm tm_grabtime = *localtime(&grab_time);
	char grab_time_buf[128];
	strftime(grab_time_buf,sizeof(grab_time_buf),"%H:%M",&tm_grabtime);
	char uptime_buf[128];
	time_t uptime_days = uptime/86400;
	time_t uptime_hours = (uptime-uptime_days*86400)/3600;
	time_t uptime_minutes = (uptime-uptime_days*86400-uptime_hours*3600)/60;
	if(uptime_days<1)
		sprintf(uptime_buf,"%d:%02d",(int)uptime_hours,(int)uptime_minutes);
	else
		sprintf(uptime_buf,"%d:%02d:%02d",(int)uptime_days,(int)uptime_hours,(int)uptime_minutes);
	
	wmove(w,0,0);
	wprintw(w,"%s  up %s, %d users,  load average: %4.2f, %4.2f, %4.2f",
	        grab_time_buf,
		uptime_buf,
		user_count,
		loadavg[0],
		loadavg[1],
		loadavg[2]
	       );
	
	wmove(w,1,0);
	wprintw(w,"%u processes: %u sleeping, %u running, %u zombie, %u stopped",
	       proccount_total,proccount_sleep,proccount_running,proccount_stopped,proccount_zombie
	       );
}

void paint_cpu(WINDOW *w, bool total_only)
{
	wclear(w);
	if(total_only) {
		wmove(w,0,0);
		wprintw(w,"CPU states:  %5.2f%% user, %5.2f%% system, %5.2f%% nice, %5.2f%% idle",
		        total_cpu_user,
		        total_cpu_sys,
		        total_cpu_nice,
		        total_cpu_idle
		       );
	} else {
		for(unsigned i=0; i<cpus; i++) {
			if(i==0) {
				wmove(w,0,0);
				wprintw(w,"CPU states: ");
			} else
				wmove(w,i,12);
			wprintw(w,"%3u", cpu[i].id);
			if(cpu[i].online)
				wprintw(w," %5.2f%% user, %5.2f%% system, %5.2f%% nice, %5.2f%% idle",
			        	cpu[i].user,
			        	cpu[i].sys,
			        	cpu[i].nice,
			        	cpu[i].idle
				       );
		}
	}
}

void paint_memory(WINDOW *w) {
	wclear(w);
	wmove(w,0,0);
	wprintw(w,"Mem:   642196K av,  313660K used,  328536K free,    2752K shrd,    8892K buff");
	wmove(w,1,0);
	wprintw(w,"Swap:  530136K av,       0K used,  530136K free                  205984K cached");
}

void paint_header(WINDOW *w) {
	int width[32];
	DistributeFields(field_def,fields,getmaxx(w)-getbegx(w),width);
	PaintHeaderFields(w,field_def,fields,width);
}

static const char *lookup_username(long userid) {
	static char buf[64];
	struct passwd *pw = getpwuid(userid);
	if(pw) {
		strncpy(buf,pw->pw_name,sizeof(buf));
		buf[sizeof(buf)-1] = '\0';
		return buf;
	} else
		return "?";
}

#define NO_USER -123
static int lookup_userid(const char *username) {
	struct passwd *pw = getpwnam(username);
	if(pw)
		return pw->pw_uid;
	else
		return NO_USER;
}

static void paint_kb(WINDOW *w, unsigned kb, int width) {
	if(kb<=99999)
		wprintw(w,"%*u",width,kb);
	else
		wprintw(w,"%*uM",width-1,kb/1024);
}

static void paint_pct(WINDOW *w, int width, double pct) {
	if(width<5)
		wprintw(w,"%*.0f",width,pct);
	else
		wprintw(w,"%*.2f",width,pct);
}

#ifdef __cplusplus
extern "C" {
#endif
static int cmp_pid(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	return p1->pid - p2->pid;
}
static int cmp_age(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	return p1->start_time - p2->start_time;
}
static int cmp_cpuusage(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	if(p1->cpu_usage < p2->cpu_usage) return -1;
	if(p1->cpu_usage > p2->cpu_usage) return  1;
	return 0;
}
static int cmp_resmem(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	if(p1->res_kb<p2->res_kb) return -1;
	if(p1->res_kb>p2->res_kb) return  1;
	return 0;
}
static int cmp_cputime(const void *pv1, const void *pv2) {
	const Process *p1 = (const Process*)pv1;
	const Process *p2 = (const Process*)pv2;
	if(p1->cpu_time<p2->cpu_time) return -1;
	if(p1->cpu_time>p2->cpu_time) return  1;
	return 0;
}
#ifdef __cplusplus
}
#endif

static void sort_processes(Sorting_t sorting) {
	switch(sorting) {
		case sorting_pid:      qsort(process,processes,sizeof(process[0]), cmp_pid);      break;
		case sorting_age:      qsort(process,processes,sizeof(process[0]), cmp_age);      break;
		case sorting_cpuusage: qsort(process,processes,sizeof(process[0]), cmp_cpuusage); break;
		case sorting_resmem:   qsort(process,processes,sizeof(process[0]), cmp_resmem);   break;
		case sorting_cputime:  qsort(process,processes,sizeof(process[0]), cmp_cputime);  break;
	}
}

void paint_processes(WINDOW *w, bool filter_idle, const char *filter_user, Sorting_t sorting) {

	sort_processes(sorting);
	
	if(filter_user && !filter_user[0]) filter_user=0;
	uid_t filter_userid;
	if(filter_user) {
		filter_userid = lookup_userid(filter_user);
	}
	
	int width[32];
	DistributeFields(field_def,fields,getmaxx(w)-getbegx(w),width);
	unsigned height = getmaxy(w)-getbegy(w);
	
	wclear(w);
	int line=0;
	for(unsigned i=0; i<processes; i++) {
		if(i>height) break;
		Process *p = process+i;
		if(filter_idle && p->is_idle) continue;
		if(filter_user && p->userid!=filter_userid) continue;
		
		const int *wi=width;
		wmove(w,line,0);
		line++;
		
		wprintw(w,"%*lu",*wi,(unsigned long)p->pid);
		waddch(w,' ');
		wi++;
		
		const char *username=lookup_username(p->userid);
		wprintw(w,"%-*.*s",*wi,*wi,username);
		waddch(w,' ');
		wi++;
		
		wprintw(w,"%*d",*wi,p->priority);
		waddch(w,' ');
		wi++;
		
		wprintw(w,"%*d",*wi,p->nice);
		waddch(w,' ');
		wi++;
		
		paint_kb(w,p->vsz_kb,*wi);
		waddch(w,' ');
		wi++;
		
		paint_kb(w,p->res_kb,*wi);
		waddch(w,' ');
		wi++;
		
		paint_kb(w,p->share_kb,*wi);
		waddch(w,' ');
		wi++;
		
		wprintw(w,"%-*.*s",*wi,*wi,"?");
		waddch(w,' ');
		wi++;
		
		paint_pct(w,*wi,p->cpu_usage);
		waddch(w,' ');
		wi++;
		
		paint_pct(w,*wi,p->mem_usage);
		waddch(w,' ');
		wi++;
		
		time_t cputime_minutes = p->cpu_time/60;
		time_t cputime_seconds = p->cpu_time%60;
		char cputime_str[16];
		sprintf(cputime_str,"%ld:%02ld", (long)cputime_minutes, (long)cputime_seconds);
		wprintw(w,"%*.*s",*wi,*wi,cputime_str);
		waddch(w,' ');
		wi++;
		
		wprintw(w,"%-*.*s",*wi,*wi,p->command);
	}
}


void do_kill(pid_t pid, int sig, WINDOW *err_window) {
	if(kill(pid,sig)!=0) {
		int save_errno=errno;
		wclear(err_window);
		wprintw(err_window,"signal(%ld,%d) failed, errno=%d, error=%s",(long)pid,sig,save_errno,strerror(save_errno));
		wgetch(err_window);
	}
}

void do_renice(pid_t pid, int value, WINDOW *err_window) {
	errno=0;
	if(setpriority(PRIO_PROCESS,pid,value)==-1 && errno!=0) {
		int save_errno=errno;
		wclear(err_window);
		wprintw(err_window,"setpriority(...,%ld,%d) failed, errno=%d, error=%s",(long)pid,value,save_errno,strerror(save_errno));
		wgetch(err_window);
	}
}
